import numpy as np

# robust imports to avoid having to update imports in tests in the future
from mumott.methods.basis_sets.nearest_neighbor import NearestNeighbor
from mumott.methods.utilities.grids_on_the_sphere import football
from mumott import ProbedCoordinates


test_probed_coordinates = ProbedCoordinates()
test_probed_coordinates.vector = np.array([[[[0.2561284, 0.35687616, 0.89835274],
                                             [0.44361984, 0.40266136, 0.80066552],
                                             [0.61035337, 0.42960518, 0.66551345]],
                                            [[0.49666735, 0.58970576, 0.63684273],
                                             [0.47785961, 0.71727289, 0.50711912],
                                             [0.44306493, 0.82084342, 0.36042967]]],
                                           [[[0.49333026, 0.24511837, 0.83459106],
                                             [0.45562494, 0.71536392, 0.52977389],
                                             [0.27418541, 0.95993642, 0.05783104]],
                                            [[0.81233662, 0.33514074, 0.47727339],
                                             [0.61245541, 0.48148378, 0.62695434],
                                             [0.36126686, 0.58749139, 0.72411333]]]])

exact_matrix = np.array([[[0.30691401, 0., 0., 0., 0.,
                           0., 0., 0., 0., 0.,
                           0., 0., 0.69308599, 0., 0.,
                           0.],
                          [0., 0., 0., 0., 0.,
                           0., 0., 0., 0., 0.,
                           0., 0., 1., 0., 0.,
                           0.]],
                         [[0., 0., 0., 0., 0.31262334,
                           0., 0.08403651, 0., 0., 0.,
                           0., 0., 0.60334015, 0., 0.,
                           0.],
                          [0., 0., 0., 0., 0.,
                           0., 0., 0., 0., 0.,
                           0., 0., 1., 0., 0.,
                           0.]]])

midpoint_matrix = np.array([[[0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.]],
                            [[0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.]]])

projection_matrix_default = np.array([[[0.30680339, 0., 0., 0., 0.,
                                        0., 0., 0., 0., 0.,
                                        0., 0., 0.69319661, 0., 0.,
                                        0.],
                                       [0., 0., 0., 0., 0.,
                                        0., 0., 0., 0., 0.,
                                        0., 0., 1., 0., 0.,
                                        0.]],
                                      [[0., 0., 0., 0., 0.31266276,
                                        0., 0.08414714, 0., 0., 0.,
                                        0., 0., 0.6031901 , 0., 0.,
                                        0.],
                                       [0., 0., 0., 0., 0.,
                                        0., 0., 0., 0., 0.,
                                        0., 0., 1., 0., 0.,
                                        0.]]])

trapz_matrix = np.array([[[0.30761719, 0., 0., 0., 0.,
                           0., 0., 0., 0., 0.,
                           0., 0., 0.69238281, 0., 0.,
                           0.],
                          [0., 0., 0., 0., 0.,
                           0., 0., 0., 0., 0.,
                           0., 0., 1., 0., 0.,
                           0.]],
                         [[0., 0., 0., 0., 0.31347656,
                           0., 0.08496094, 0., 0., 0.,
                           0., 0., 0.6015625 , 0., 0.,
                           0.],
                          [0., 0., 0., 0., 0.,
                           0., 0., 0., 0., 0.,
                           0., 0., 1., 0., 0.,
                           0.]]])

romb_matrix = np.array([[[0.30679034, 0., 0., 0., 0.,
                          0., 0., 0., 0., 0.,
                          0., 0., 0.69320966, 0., 0.,
                          0.],
                         [0., 0., 0., 0., 0.,
                          0., 0., 0., 0., 0.,
                          0., 0., 1., 0., 0.,
                          0.]],
                        [[0., 0., 0., 0., 0.31264872,
                          0., 0.08413408, 0., 0., 0.,
                          0., 0., 0.6032172 , 0., 0.,
                          0.],
                         [0., 0., 0., 0., 0.,
                          0., 0., 0., 0., 0.,
                          0., 0., 1., 0., 0.,
                          0.]]])


def test_nearest_neighbor():
    directions = football()
    basis_set = NearestNeighbor(directions, test_probed_coordinates,
                                integration_tolerance=1e-5, integration_maxiter=20)
    print(np.array_repr(basis_set.projection_matrix))
    assert np.allclose(basis_set.projection_matrix, exact_matrix)
    basis_set = NearestNeighbor(directions, test_probed_coordinates, integration_mode='midpoint')
    print(np.array_repr(basis_set.projection_matrix))
    assert np.allclose(basis_set.projection_matrix, midpoint_matrix)
    basis_set = NearestNeighbor(directions, test_probed_coordinates)
    print(np.array_repr(basis_set.projection_matrix))
    assert np.allclose(basis_set.projection_matrix, projection_matrix_default)
    # these should be very similar to the default
    basis_set = NearestNeighbor(directions, test_probed_coordinates, integration_mode='trapezoid')
    print(np.array_repr(basis_set.projection_matrix))
    assert np.allclose(basis_set.projection_matrix, trapz_matrix)
    basis_set = NearestNeighbor(directions, test_probed_coordinates, integration_mode='romberg')
    print(np.array_repr(basis_set.projection_matrix))
    assert np.allclose(basis_set.projection_matrix, romb_matrix)
