import numpy as np

from mumott import DataContainer
from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.residual_calculators import ZHTTResidualCalculator
from mumott.methods.projectors import SAXSProjector

rotated_coefficients_test = np.array([1., 0.55758988, 0.66263709, -0.06211013, 0.42547438,
                                      -0.25518522, -0.2805971, 0.09502369, 0.37557435, -0.28929866,
                                      -0.34688224, -0.18575653, -0.17188444, -0.66661516, -0.24234923])
gradient_test = np.array([-45.23868716, 13.43821073, 17.86106824, 25.96389411,
                          31.58224747])
residual_test = np.array([[[[14.88301225, 13.19567949, 15.76450625],
                           [13.46301225, 13.32567949, 17.31450625],
                           [12.62301225, 13.70567949, 18.83450625],
                           [12.41301225, 14.14567949, 19.81450625]],
                          [[16.69301225, 12.99567949, 15.65450625],
                           [15.39301225, 13.98567949, 17.48450625],
                           [13.93301225, 14.74567949, 19.34450625],
                           [13.41301225, 15.38567949, 20.35450625]],
                          [[18.95301225, 14.27567949, 14.32450625],
                           [15.83301225, 16.14567949, 16.20450625],
                           [15.27301225, 18.04567949, 15.39450625],
                           [14.76301225, 20.00567949, 15.52450625]],
                          [[19.14301225, 15.72567949, 14.49450625],
                           [17.26301225, 18.03567949, 14.60450625],
                           [16.02301225, 21.04567949, 14.46450625],
                           [16.42301225, 23.09567949, 14.03450625]]]])


def test_zonal_harmonics_gradient_calculator():

    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics(ell_max=4)
    dc.geometry.integration_samples = 1
    pr = SAXSProjector(dc.geometry)
    zh = ZHTTResidualCalculator(dc, bs, pr)

    zh._theta[...] = 1
    zh._phi[...] = 1

    print(hex(hash(zh))[2:8])
    assert hex(hash(zh))[2:8] == '756ba7'

    coeffs_test = np.ones((*pr.volume_shape, 3+2))
    zh.coefficients = coeffs_test
    print(np.array_repr(zh.rotated_coefficients[0, 0, 0, :]))
    assert np.allclose(zh.rotated_coefficients[0, 0, 0, :], rotated_coefficients_test)

    result = zh.get_residuals(get_gradient=True, get_weights=True)
    print(np.array_repr(result['gradient'][0, 0, 0, :]))
    assert np.allclose(result['gradient'][0, 0, 0, :], gradient_test)
    print(np.array_repr(result['residuals']))
    assert np.allclose(result['residuals'], residual_test)

    result = zh.get_gradient()
    assert np.allclose(result['gradient'][0, 0, 0, :], gradient_test)

    bs.ell_max == 6
    assert zh.is_dirty
    zh._update()
    assert not zh.is_dirty
    assert np.allclose(zh._zonal_coefficients[:, :, :, 3:5], 0)


def test_str():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics(ell_max=4)
    pr = SAXSProjector(dc.geometry)
    zh = ZHTTResidualCalculator(dc, bs, pr)
    s = str(zh)
    print(s)
    # NB: Object has random initial state, uses ProbedCoordinates hash
    assert 'dc31bb' in s


def test_html():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics(ell_max=4)
    pr = SAXSProjector(dc.geometry)
    zh = ZHTTResidualCalculator(dc, bs, pr)
    s = zh._repr_html_()
    assert 'dc31bb' in s
    assert 'dirty' in s
