import numpy as np
from mumott import DataContainer

from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjector
from mumott.methods.residual_calculators import ZHTTResidualCalculator
from mumott.optimization.loss_functions import SquaredLoss
from mumott.optimization.optimizers.zonal_harmonics_optimizer import ZHTTOptimizer
from mumott.optimization.regularizers import L2Norm

expected_solution = np.array([-11.75632, -5.23421553, -0.1452865, 3.14159265])


def test_ZHTT_optimizer():

    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics(ell_max=2)
    pr = SAXSProjector(dc.geometry)
    meth = ZHTTResidualCalculator(dc, bs, pr)
    lf = SquaredLoss(meth)
    regularizer = L2Norm()
    lf.add_regularizer(name='l2',
                       regularizer=regularizer,
                       regularization_weight=1e4)

    x0 = np.zeros((*dc.geometry.volume_shape, 4))
    opt = ZHTTOptimizer(lf, x0, maxiter=2)
    opt._step_size_parameter = 0.005
    results = opt.optimize()
    print(repr(results['x'][0, 0, 0, :]))
    assert np.allclose(results['x'][0, 0, 0, :], expected_solution)


def test_hash():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics(ell_max=2)
    pr = SAXSProjector(dc.geometry)
    meth = ZHTTResidualCalculator(dc, bs, pr)
    lf = SquaredLoss(meth)
    regularizer = L2Norm()
    lf.add_regularizer(name='l2',
                       regularizer=regularizer,
                       regularization_weight=1e4)

    x0 = np.zeros((*dc.geometry.volume_shape, 4))
    opt = ZHTTOptimizer(lf, x0, maxiter=2, step_size_parameter=0.001)
    lf.get_loss(x0, gradient_part='full')
    assert hex(hash(opt))[2:8] == '1ddf66'
