import pytest  # noqa
import numba
import numpy as np
from mumott.core.cuda_kernels import (cuda_weighted_difference, cuda_scaled_difference,
                                      cuda_sum, cuda_rescale, cuda_difference,
                                      cuda_rescale_array,
                                      cuda_l1_gradient, cuda_tv_gradient,
                                      cuda_framewise_contraction,
                                      cuda_framewise_contraction_adjoint)

test_array_1 = np.arange(256.).reshape(4, 4, 4, 4).astype(np.float32)
test_array_2 = np.arange(0, 128., 0.5).reshape(4, 4, 4, 4).astype(np.float32)
test_array_3 = np.arange(-64., 64., 0.5).reshape(4, 4, 4, 4).astype(np.float32)
test_array_4 = np.arange(0, 4 * 4 * 4 * 6).reshape(4, 4, 4, 6).astype(np.float32)
test_matrix_stack = np.arange(0, 4 * 6 * 4).reshape(4, 6, 4).astype(np.float32)

out_1 = np.array([[-1280., -1328.25, -1377., -1426.25],
                  [-1476., -1526.25, -1577., -1628.25],
                  [-1680., -1732.25, -1785., -1838.25],
                  [-1892., -1946.25, -2001., -2056.25]]).astype(np.float32)


def test_cuda_weighted_difference():
    kernel = cuda_weighted_difference(test_array_1.shape)
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(test_array_2)
    a3 = numba.cuda.to_device(test_array_3)
    kernel(a1, a2, a3)
    print(repr(np.array(a2)[2, 2]))
    assert np.allclose(np.array(a2)[2, 2], out_1)


out_2 = np.array([[-2480., -2576., -2673., -2771.],
                  [-2870., -2970., -3071., -3173.],
                  [-3276., -3380., -3485., -3591.],
                  [-3698., -3806., -3915., -4025.]], dtype=np.float32)


def test_cuda_scaled_difference():
    kernel = cuda_scaled_difference(test_array_1.shape)
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(test_array_2)
    a3 = numba.cuda.to_device(test_array_3)
    kernel(a1, a2, a3)
    print(repr(np.array(a2)[2, 2]))
    assert np.allclose(np.array(a2)[2, 2], out_2)


out_3 = np.array([[240., 241.5, 243., 244.5],
                 [246., 247.5, 249., 250.5],
                 [252., 253.5, 255., 256.5],
                 [258., 259.5, 261., 262.5]], dtype=np.float32)


def test_cuda_sum():
    kernel = cuda_sum(test_array_1.shape)
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(test_array_2)
    kernel(a1, a2)
    print(repr(np.array(a1)[2, 2]))
    assert np.allclose(np.array(a1)[2, 2], out_3)


out_4 = np.array([[80., 80.5, 81., 81.5],
                  [82., 82.5, 83., 83.5],
                  [84., 84.5, 85., 85.5],
                  [86., 86.5, 87., 87.5]], dtype=np.float32)


def test_cuda_difference():
    kernel = cuda_difference(test_array_1.shape)
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(test_array_2)
    kernel(a1, a2)
    print(repr(np.array(a1)[2, 2]))
    assert np.allclose(np.array(a1)[2, 2], out_4)


out_5 = np.array([[12800., 12960.5, 13122., 13284.5],
                  [13448., 13612.5, 13778., 13944.5],
                  [14112., 14280.5, 14450., 14620.5],
                  [14792., 14964.5, 15138., 15312.5]], dtype=np.float32)


def test_cuda_rescale_array():
    kernel = cuda_rescale_array(test_array_1.shape)
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(test_array_2)
    kernel(a1, a2)
    print(repr(np.array(a1)[2, 2]))
    assert np.allclose(np.array(a1)[2, 2], out_5)


out_6 = np.array([[144., 144.9, 145.8, 146.7],
                  [147.59999, 148.5, 149.4, 150.3],
                  [151.2, 152.09999, 153., 153.9],
                  [154.8, 155.7, 156.59999, 157.5]], dtype=np.float32)


def test_cuda_rescale():
    kernel = cuda_rescale(test_array_1.shape)
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(test_array_2)
    kernel(a1, a2)
    print(repr(np.array(a1)[2, 2]))
    assert np.allclose(np.array(a1)[2, 2], out_6)


out_7 = np.array([[80.0001, 80.5001, 81.0001, 81.5001],
                  [82.0001, 82.5001, 83.0001, 83.5001],
                  [84.0001, 84.5001, 85.0001, 85.5001],
                  [86.0001, 86.5001, 87.0001, 87.5001]], dtype=np.float32)


def test_cuda_l1_gradient():
    kernel = cuda_l1_gradient(test_array_1.shape)
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(test_array_2)
    kernel(a1, a2)
    print(repr(np.array(a2)[2, 2]))
    assert np.allclose(np.array(a2)[2, 2], out_7)


out_8 = np.array([[0., 0., 0., 0.],
                  [82., 82.5, 83., 83.5],
                  [84., 84.5, 85., 85.5],
                  [0., 0., 0., 0.]], dtype=np.float32)


def test_cuda_tv_gradient():
    kernel = cuda_tv_gradient(test_array_1.shape)
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(test_array_2)
    kernel(a1, a2)
    print(repr(np.array(a2)[2, 2]))
    assert np.allclose(np.array(a2)[2, 2], out_8)


out_9 = np.array([[31982., 34566., 37150., 39734., 42318., 44902.],
                  [32774., 35422., 38070., 40718., 43366., 46014.],
                  [33566., 36278., 38990., 41702., 44414., 47126.],
                  [34358., 37134., 39910., 42686., 45462., 48238.]],
                 dtype=np.float32)


def test_framewise_contraction():
    kernel = cuda_framewise_contraction(test_array_1.shape, *test_matrix_stack.shape[1:])
    a1 = numba.cuda.to_device(test_array_1)
    a2 = numba.cuda.to_device(np.zeros_like(test_array_4))
    m = numba.cuda.to_device(test_matrix_stack)
    kernel(a1, m, a2)
    print(repr(np.array(a2)[2, 2]))
    assert np.allclose(np.array(a2)[2, 2], out_9)


out_10 = np.array([[84460., 85915., 87370., 88825.],
                   [86548., 88039., 89530., 91021.],
                   [88636., 90163., 91690., 93217.],
                   [90724., 92287., 93850., 95413.]],
                  dtype=np.float32)


def test_framewise_contraction_adjoint():
    kernel = cuda_framewise_contraction_adjoint(test_array_1.shape, *test_matrix_stack.shape[1:])
    a1 = numba.cuda.to_device(test_array_4)
    a2 = numba.cuda.to_device(np.zeros_like(test_array_1))
    m = numba.cuda.to_device(test_matrix_stack)
    kernel(a1, m, a2)
    print(repr(np.array(a2)[2, 2]))
    assert np.allclose(np.array(a2)[2, 2], out_10)
